import ldap

from .. import auth

# Required when using self-signed certificates.
ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_NEVER)


class Driver:
    def __init__(self, config):
        self.config = config

    def session(self):
        return Session(**self.config)


class Session:
    def __init__(self, domain, search, server):
        self.domain = domain
        self.search = search

        # Cache group membership to avoid repeated lookups.
        self.groups = {}

        # Connect, and ensure the connection is encrypted.
        self.client = ldap.initialize(server)
        self.client.start_tls_s()

    def authenticate(self, username, password):
        # Authenticate as the user.
        self.client.simple_bind_s(self.domain + "\\" + username, password)

        # Retrieve user attributes.
        result = self.lookup("user", username)

        if not result:
            raise Exception("search: no results")

        # Each result is a tuple: (distinguishedName, attributes)
        attrs = result[1]
        props = {
            "email":      "mail",
            "first_name": "givenName",
            "last_name":  "sn",
        }

        # Map from LDAP attributes to account properties.
        for prop, attr in props.items():
            props[prop] = ""

            if attr in attrs and attrs[attr]:
                props[prop] = attrs[attr][0].decode("utf8")

        return auth.Account(**props)

    def close(self):
        self.client.unbind_ext_s()

    def lookup(self, kind, *args):
        results = self.client.search_s(
            self.search[kind]["base"], ldap.SCOPE_SUBTREE,
            self.search[kind]["filter"] % tuple(args),
        )

        return next(iter(results), None)

    def member(self, username, groups):
        for group in groups:
            # Populate the cache for this group, if necessary.
            if group not in self.groups:
                self.groups[group] = self.member_query(username, group)

            if self.groups[group]:
                return True

        return False

    def member_query(self, username, group):
        # Get the DN of the group.
        group = self.lookup("group", group)

        # Get membership.
        if group and self.lookup("member", username, group[0]):
            return True

        return False



import django.apps
import django.conf
import importlib
import inspect


class AppConfig(django.apps.AppConfig):
    name = "netbox_rbac"

    def ready(self):
        # Patch the restricted queryset manager to remove view based restrictions.
        import utilities.querysets

        setattr(utilities.querysets.RestrictedQuerySet, "restrict", lambda qs, *args: qs)

        # Patch views.
        from .mixins import (
            BulkDeleteView,
            BulkEditView,
            BulkRenameView,
            ObjectDeleteView,
            ObjectEditView,
        )

        replace = {
            "BulkDeleteView":   BulkDeleteView,
            "BulkEditView":     BulkEditView,
            "BulkRenameView":   BulkRenameView,
            "ObjectDeleteView": ObjectDeleteView,
            "ObjectEditView":   ObjectEditView,
        }

        # Core applications.
        apps = ["circuits", "dcim", "extras", "ipam", "tenancy", "virtualization"]
        conf = django.conf.settings.RBAC

        # Custom applications? Patch those too.
        if "APPS" in conf:
            apps += conf["APPS"]

        # For each view in each app, replace various generic view classes.
        for app in apps:
            module = importlib.import_module("%s.views" % app)

            for _, obj in inspect.getmembers(module, inspect.isclass):
                if obj.__module__ == module.__name__:
                    bases = list(obj.__bases__)

                    for i, base in enumerate(bases):
                        for name, view in replace.items():
                            if name == base.__name__:
                                bases[i] = view

                    obj.__bases__ = tuple(bases)



import functools
import threading

from django.core.exceptions   import PermissionDenied
from django.db.models.signals import m2m_changed, pre_save


# Ignore changes related to authentication.
ignore_modules = [
    "django.contrib.auth.models",
    "django.contrib.sessions.models",
    "netbox_rbac.models",
    "users.models",
]

# Track the current request so rules can evaluate request attributes.
requests = {}


def request():
    return requests.get(threading.current_thread())


class Middleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        handler = functools.partial(self.has_perm, request)
        signals = (m2m_changed, pre_save)

        requests[threading.current_thread()] = request

        for signal in signals:
            signal.connect(handler)

        response = self.get_response(request)

        for signal in signals:
            signal.disconnect(handler)

        del requests[threading.current_thread()]

        return response

    def has_perm(self, request, instance, **kwargs):
        if instance.__module__ in ignore_modules:
            return

        # Determining the correct permission is surprisingly tricky.
        #
        # Operations performed via the REST API follow the typical REST method
        # convention, so we borrow the mapping from the Django REST framework.
        #
        # Operations performed via the web UI are less consistent, and mostly
        # use the POST method, so we have to see if the object already has a
        # primary key in order to distinguish between 'add' and 'change'.
        method_operation = {
            "DELETE": "delete",
            "GET":    "view",
            "HEAD":   "view",
            "PATCH":  "change",
            "POST":   "add",
            "PUT":    "change",
        }

        oper = "change" if instance.pk else method_operation[request.method]
        perm = "%s.%s_%s" % (instance._meta.app_label, oper, instance._meta.model_name,)

        if not request.user.has_perm(perm, instance):
            raise PermissionDenied("%s %s" % (oper, instance))



from netbox_rbac import macros

from netbox_rbac.tests.types import *

r1 = Rack(name="R1", site=None)
r2 = Rack(name="R2", site=Site(name="DC1"))


class TestMacros:
    def test_get(self):
        tests = [
            (r1, "name", "R1"),
            (r1, "none", None),
            (r1, "site", None),
            (r1, "site.name", None),
            (r2, "site.name", "DC1"),
        ]

        for obj, path, value in tests:
            assert macros.get(obj, path) == value

    def test_match(self):
        tests = [
            (r1, "name", "R1", True),
            (r1, "name", "R2", False),
            (r1, "site.name", "DC1", False),
            (r2, "site.name", "DC1", True),
            (r1, "site.name", "DC2", False),
            (r2, "site.name", "DC2", False),
            (r2, "site.name", "DC*", True),
            (r2, "site.name", ["DC1", "DC2"], True),
        ]

        for obj, path, value, result in tests:
            assert macros.match(obj, path, value) == result

    def test_match_or_none(self):
        tests = [
            (r1, "name", "R1", True),
            (r1, "name", "R2", False),
            (r1, "site.name", "DC1", True),
            (r2, "site.name", "DC1", True),
            (r1, "site.name", "DC2", True),
            (r2, "site.name", "DC2", False),
            (r2, "site.name", "DC*", True),
            (r2, "site.name", ["DC1", "DC2"], True),
        ]

        for obj, path, value, result in tests:
            assert macros.match_or_none(obj, path, value) == result



from wcmatch.fnmatch import fnmatch


def get(obj, path):
    for attr in path.split("."):
        if not hasattr(obj, attr):
            return None

        obj = getattr(obj, attr)

    return obj


def match(obj, *args):
    *paths, values = args

    return _walk_match(obj, paths, values, False)


def match_or_none(obj, *args):
    *paths, values = args

    return _walk_match(obj, paths, values, True)


def _walk_match(obj, paths, values, default):
    for path in paths:
        value = get(obj, path)

        if value is not None:
            return fnmatch(value, values)

    return default



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fh:
    packages = fh.read().splitlines()

setuptools.setup(
    name="netbox-rbac",
    version="1.0.17",
    author="Eric Busto",
    author_email="ebusto@nvidia.com",
    description="Opinionated RBAC for NetBox",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ebusto/netbox-rbac",
    packages=setuptools.find_packages(),
    install_requires=packages,
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],
)



from django.contrib.auth.models     import User
from django.contrib.postgres.fields import ArrayField

from django.db import models


class Profile(models.Model):
    user = models.OneToOneField(User, on_delete=models.CASCADE)

    roles = ArrayField(models.CharField(max_length=255), blank=True, default=list,)

    class Meta:
        managed = True
